library spea_mobile_socket_conn;

import 'dart:io';
import 'package:connectivity/connectivity.dart';
import 'package:flutter/cupertino.dart';
import 'package:flutter/material.dart';
import 'package:shared_preferences/shared_preferences.dart';
import 'package:wifi_configuration/wifi_configuration.dart';
import 'package:spea_mobile_qrcode/spea_mobile_qrcode.dart';

///Add this to info.plist
///
///<key>io.flutter.embedded_views_preview</key>
///<true/>


class SpeaMobileClient {

  static const String kSP_PROFILES_KEY = "Profiles";

  //Show the list of profiles
  static Future<dynamic> showProfileList(
      {
        @required BuildContext context ,
        @required String appId ,
        @required Function doneHandler ,
        @required Function errorHandler ,
        @required Function dataHandler
      }
      ) async {

    print("Showing profile list ...");

    dynamic res = await showDialog(
        context: context,
        builder: (context){
          return _ProfileList(appId: appId, doneHandler: doneHandler, dataHandler: dataHandler, errorHandler: errorHandler,);
        }
    );
    return res;
  }
}

class _ProfileList extends StatefulWidget {
  final String appId;
  final Function dataHandler;
  final Function errorHandler;
  final Function doneHandler;

  _ProfileList(
      {
        @required this.appId,
        @required this.dataHandler,
        @required this.doneHandler,
        @required this.errorHandler
      }
      );

  @override
  __ProfileListState createState() => __ProfileListState();
}

class __ProfileListState extends State<_ProfileList> {
  bool _loading;
  List<Widget> profiles;
  String _error;

  bool _connected;

  Socket _resultSocket;

  @override
  void initState() {
    print("Init state of profile list");
    _connected = false;
    _loading = false;
    _error = "";
    _resultSocket = null;
    super.initState();
  }

  @override
  Widget build(BuildContext context) {

    if (_loading){
      print("Loading ...");
      return Center(child: CircularProgressIndicator(),);
    }

    if (_error != ""){
      print("Error ...");
      return AlertDialog(
        backgroundColor: Colors.white,
        content: FlatButton.icon(
          onPressed: (){Navigator.pop(context);},
          icon: Icon(Icons.error, color: Colors.red,),
          label: Flexible(child: Text(_error, style: TextStyle(color: Colors.red, fontSize: 16.0),)),),
      );
    }


    if (_connected){
      Future.delayed(Duration(seconds: 1)).then((niente){
        try {
          Navigator.pop(context, _resultSocket);
        } catch (e) {
          print("L' utente ha gia cancellato il dialogo");
        }
      });
      return AlertDialog(
        backgroundColor: Colors.white,
        content: FlatButton.icon(
          onPressed: (){ Navigator.pop(context, _resultSocket); },
          icon: Icon(Icons.done, color: Colors.green,),
          label: Text("Connected!", style: TextStyle(color: Colors.green, fontSize: 16.0),),),
      );
    }

    return FutureBuilder (
        future: fillOptions ( ) ,
        builder: ( context , snapshot ){
          switch (snapshot.connectionState) {

            case ConnectionState.waiting:
              print("Loading options ... ");
              return Center( child: new CircularProgressIndicator( ) , );

            default:
              if ( snapshot.hasError ) {
                print("Error loading profiles ... ");
                return AlertDialog(
                  content: FlatButton.icon(
                    onPressed: (){Navigator.pop(context);},
                    icon: Icon(Icons.error_outline, color: Colors.red,),
                    label: Text("Error reading profiles", style: TextStyle(color: Colors.red, fontSize: 16.0 ),),
                  ),
                );
              } else {
                profiles = snapshot.data;
                print("Loaded ${profiles.length} profiles ...");
                return SimpleDialog (
                  children: profiles ,
                );
              }
          }
        }
    );
  }

  static void askProfileNameAndSave(Profile profile, BuildContext context) {

    print("Saving new profile ...");
    
    TextEditingController textController = new TextEditingController();

    showDialog(
        context: context,
        builder: (context){
          return AlertDialog(
            title: Text('Save profile'),
            content:TextField(
              controller: textController,
              decoration: InputDecoration(
                hintText: "Insert profile title",
                border: OutlineInputBorder(),
              ),
            ),
            actions: <Widget>[
              FlatButton(
                child: Text('Cancel'),
                onPressed: () {
                  Navigator.of(context).pop();
                },
              ),
              FlatButton(
                child: Text('Save'),
                onPressed: () async {
                  var sp = await SharedPreferences.getInstance();
                  sp.setString(textController.text, profile.asString());
                  List<String> tmp = sp.getStringList(SpeaMobileClient.kSP_PROFILES_KEY) ?? List<String>();
                  tmp.add(textController.text.trim());
                  sp.setStringList(SpeaMobileClient.kSP_PROFILES_KEY, tmp);
                  Navigator.pop(context);
                  context.rootAncestorStateOfType(TypeMatcher<_ProfileList>()).setState((){});
                },
              ),
            ],
          );
        }
    );
  }

  Widget getAddProfileOption(context) {

    return SimpleDialogOption(
      onPressed: () async {
        String result = await openScanner(context);
        if (result != QrCodeViewState.QRcancelMsg){
          Profile scannedProfile = Profile.fromString(result);
          Navigator.pop(context);
          askProfileNameAndSave(scannedProfile, context);
          //Reaload UI
          setState( () {} );
        }
      },
      child: Row(
        children: <Widget>[
          Icon(Icons.add, size: 25 ,),
          Text("Scan QR to add a new Profile"),
        ],
      ),
    );
  }

  Future<List<Widget>> fillOptions() async {

    SharedPreferences shared = await SharedPreferences.getInstance();

    //Get all profile titles
    List<String> profileKeys = shared.getStringList(SpeaMobileClient.kSP_PROFILES_KEY) ?? List<String>();
    List<Widget> options = List();

    //For each profile create a dialog option
    profileKeys.forEach ( ( profileTitle ) {
      options.add ( SimpleDialogOption (
          onPressed: ( ) async {
            var sp = await SharedPreferences.getInstance ( );
            Profile profile = Profile.fromString ( sp.getString ( profileTitle ) );
            tryConnectToProfile(profile);
          } ,
          child: Row (
            mainAxisAlignment: MainAxisAlignment.spaceBetween ,
            crossAxisAlignment: CrossAxisAlignment.center ,
            mainAxisSize: MainAxisSize.max ,
            children: <Widget>[
              ///PROFILE TITLE
              Row (
                children: <Widget>[
                  Icon ( Icons.chevron_right ) ,
                  Text (
                    profileTitle ?? "" , style: TextStyle ( fontSize: 18 ) , ) ,
                ] ,
              ) ,
              ///DELETE BUTTON
              IconButton (
                icon: Icon ( Icons.delete ) ,
                onPressed: ( ) async {

                  ///On delete
                  var sp = await SharedPreferences.getInstance ( );

                  ///remove data saved in shared prefs
                  sp.remove ( profileTitle );

                  ///remove the key from the key list saved in shared prefs
                  var profileKeys = sp.getStringList (
                      SpeaMobileClient.kSP_PROFILES_KEY );
                  profileKeys.remove ( profileTitle );
                  sp.setStringList (
                      SpeaMobileClient.kSP_PROFILES_KEY , profileKeys );
                } ,
              ) ,
            ] ,
          )
      ) );
    } );

    //Insert add profile dialog option
    options.add( getAddProfileOption(context) );

    return options;
  }

  Future tryConnectToProfile(Profile profile) async {

    setState(() {
      _loading = true;
      _error = "";
    });

    //Connect to wifi with scanned configuration
    String res = await _connectToWifi(ssid: profile.mappa["SSID"], pwd: profile.mappa["PWD"]);

    if (res == "connected" || res == "alreadyConnected")
      while(( await ( Connectivity().checkConnectivity() ) ) != ConnectivityResult.wifi) {}
    else{
      setState(() {
        _loading = false;
        _error = "Cannot connect to wifi network";
      });
    }

    connectToServer(
      profile.mappa["IP"],
      int.tryParse(profile.mappa["PORT"]),
    );

  }

  Future<String> _connectToWifi({@required String ssid, @required String pwd}) async {
    String connectionState = "";

    connectionState = await WifiConfiguration.connectToWifi(ssid, pwd, widget.appId);

    print("Wifi connect attempt: $connectionState");
    switch (connectionState) {
      case "connected":
        break;
    //in case of wifi successfully connected
      case "alreadyConnected":
        break;
    //in case wifi already connected
    //Only returned in ios
      case "notConnected":
        break;
    //in case when wifi not availabel or not able to connect to specific wifi
    //when not connected status returned in android user will be directed to wifi
    // settings of device to select the exact wifi to connect.
      case "profileAlreadyInstalled":
        break;
    //in case of wifi configuration profile is installed already
    //Only returned in ios
      case "locationNotAllowed":
        break;
    //in case location permission rejected
    //Only returned in android
    }
    return connectionState;
  }

  connectToServer(String ip, int port,) async {

    //Try to connect to specified ip-address
    Socket.connect(ip, port, timeout: Duration(seconds: 4)).then(
          (socket){
        print("Connecting to ${socket.remoteAddress} using ${socket.address}");
        //Add handler/listener
        socket.listen(masterDataHandler, onDone: masterDoneHandler, onError: masterErrorHandler,);
        setState(() {
          _loading = false;
          _error = "";
          _connected = true;
          _resultSocket = socket;
        });
      },
      onError: (error){
        masterErrorHandler(error.toString());
        print(error.toString() ?? "Unkown error!");
      },
    ).catchError( (error){
      print("Error trying connecting to server");
      masterErrorHandler(error.toString());
    });
  }

  masterDataHandler(List<int> data){
    widget.dataHandler(data);
  }

  masterDoneHandler(){
    widget.doneHandler();
  }

  masterErrorHandler(dynamic error){
    if (mounted)
      setState(() {
        _loading = false;
        _error = error;
      });
    widget.errorHandler(error);
  }
}

class Profile {

  Map<String,String> mappa;

  Profile(this.mappa);

  Profile.fromString(String fullStr){
    print("From String Constructor $fullStr");
    mappa = new Map<String,String>();
    fullStr.split("\n").forEach((str){
      List<String> tmp = str.split("=");
      if (tmp.length > 1)
        mappa[tmp[0]] = tmp[1];
    });
  }

  String asString(){
    String result = "";

    mappa.forEach((key,value){
      result += "$key=$value\n";
    });

    result.replaceRange(result.lastIndexOf("\n"), result.length, "");

    print("Converted profile to String: $result");

    return result;
  }

}





